import numpy as np
from scipy import linalg
from .kernels import Kernel
from .obj import CalAicObj, CalMultiObj


def _compute_betas_gwr(y, x, wi):
    """
    compute MLE coefficients using iwls routine

    Methods: p189, Iteratively (Re)weighted Least Squares (IWLS),
    Fotheringham, A. S., Brunsdon, C., & Charlton, M. (2002).
    Geographically weighted regression: the analysis of spatially varying relationships.
    """
    xt = (x * wi).T
    xtx = np.dot(xt, x)
    xtx_inv_xt = linalg.solve(xtx, xt)
    betas = np.dot(xtx_inv_xt, y)
    return betas, xtx_inv_xt


class GTWR:
    """
    Geographically and Temporally Weighted Regression
    """

    def __init__(
            self,
            coords: np.ndarray,
            t: np.ndarray,
            X: np.ndarray,
            y: np.ndarray,
            bw: float,
            tau: float,
            kernel: str = 'bisquare',
            fixed: bool = True,
            constant: bool = True
    ):
        """
        Parameters
        ----------
        coords        : array-like
                        n*2, collection of n sets of (x,y) coordinates of
                        observations

        t             : array
                        n*1, time location

        y             : array
                        n*1, dependent variable

        X             : array
                        n*k, independent variable, excluding the constant

        bw            : scalar
                        bandwidth value consisting of either a distance or N
                        nearest neighbors; user specified or obtained using
                        sel

        tau           : scalar
                        spatio-temporal scale

        kernel        : string
                        type of kernel function used to weight observations;
                        available options:
                        'gaussian'
                        'bisquare'
                        'exponential'

        fixed         : boolean
                        True for distance based kernel function and  False for
                        adaptive (nearest neighbor) kernel function (default)

        constant      : boolean
                        True to include intercept (default) in model and False to exclude
                        intercept.

        after call fit() method:
        Attributes
        ----------
        obj.betas           : array
                              n*k, estimated coefficients

        obj.pre             : array
                              n*1, predicted y values

        obj.CCT             : array
                              n*k, scaled variance-covariance matrix

        obj.df_model        : integer
                              model degrees of freedom

        obj.df_reside       : array
                              n*1, residuals of the response

        obj.RSS             : scalar
                              residual sum of squares

        obj.CCT             : array
                              n*k, scaled variance-covariance matrix

        obj.ENP             : scalar
                              effective number of parameters, which depends on
                              obj.sigma2

        obj.tr_S            : float
                              trace of S (hat) matrix

        obj.tr_STS          : float
                              trace of STS matrix

        obj.R2              : float
                              R-squared for the entire model (1- RSS/TSS)

        obj.adj_R2          : float
                              adjusted R-squared for the entire model

        obj.aic             : float
                              Akaike information criterion

        obj.aic_c           : float
                              corrected Akaike information criterion to account
                              to account for model complexity (smaller
                              bandwidths)

        obj.bic             : float
                              Bayesian information criterio

        obj.sigma2          : float
                              sigma squared (residual variance) that has been
                              corrected to account for the ENP

        obj.std_res         : array
                              n*1, standardised residuals

        obj.bse             : array
                              n*k, standard errors of parameters (betas)

        obj.diagonal        : array
                              n*1, leading diagonal of S matrix

        obj.CooksD          : array
                              n*1, Cook's D

        obj.t_values        : array
                              n*k, local t-statistics

        obj.llf             : scalar
                              log-likelihood of the full model; see
                              pysal.contrib.glm.family for damily-sepcific
                              log-likelihoods

        Examples
        --------
        import numpy as np
        from mgtwr.model import GTWR
        np.random.seed(1)
        u = np.array([(i -1)%12 for i in range(1,1729)]).reshape(-1,1)
        v = np.array([((i-1)%144)//12 for i in range(1,1729)]).reshape(-1,1)
        t = np.array([(i-1)//144 for i in range(1,1729)]).reshape(-1,1)
        x1 = np.random.uniform(0,1,(1728,1))
        x2 = np.random.uniform(0,1,(1728,1))
        epsilon = np.random.randn(1728,1)
        beta0 = 5
        beta1 = 3 + (u + v + t)/6
        beta2 = 3+((36-(6-u)**2)*(36-(6-v)**2)*(36-(6-t)**2))/128
        y = beta0 + beta1 * x1 + beta2 * x2 + epsilon
        coords = np.hstack([u,v])
        X = np.hstack([x1,x2])
        gtwr = GTWR(coords, t, X, y, 0.8, 1.1, kernel='gaussian', fixed=True)
        gtwr.fit()
        print(gtwr.R2)
        0.9834515760708834
        """
        self.coords = coords
        self.t = t
        self.y = y
        self.n = X.shape[0]
        self.bw = bw
        self.tau = tau
        self.kernel = kernel
        self.fixed = fixed
        self.constant = constant
        if self.constant:
            self.X = np.hstack([np.ones((self.n, 1)), X])
        else:
            self.X = X
        self.k = self.X.shape[1]
        self.bw_s = self.bw
        self.bw_t = np.sqrt(self.bw ** 2 / self.tau)
        self.diagonal = None
        self.pre = None
        self.reside = None
        self.betas = None
        self.tr_S = None
        self.ENP = None
        self.tr_STS = None
        self.TSS = None
        self.RSS = None
        self.sigma2 = None
        self.CCT = None
        self.std_res = None
        self.bse = None
        self.cooksD = None
        self.t_values = None
        self.df_model = None
        self.df_reside = None
        self.R2 = None
        self.adj_R2 = None
        self.llf = None
        self.aic = None
        self.aic_c = None
        self.bic = None

    def fit(self):
        """
        fit GTWR models
        """
        diagonal = np.empty((self.n, 1))
        reside = np.empty((self.n, 1))
        pre = np.empty((self.n, 1))
        betas, CCT = np.empty((self.n, self.k)), np.empty((self.n, self.k))
        tr_STS = 0
        for i in range(self.n):
            diagonal[i] = self._local_fit(i)[0]
            reside[i] = self._local_fit(i)[1]
            pre[i] = self._local_fit(i)[2]
            betas[i] = self._local_fit(i)[3]
            CCT[i] = self._local_fit(i)[4]
            tr_STS += self._local_fit(i)[5]
        self.diagonal = diagonal
        self.reside = reside
        self.pre = pre
        self.betas = betas
        self.ENP = np.sum(diagonal)
        self.tr_S = self.ENP
        self.tr_STS = tr_STS
        self.TSS = np.sum((self.y - np.mean(self.y)) ** 2)
        self.RSS = np.sum(reside ** 2)
        self.sigma2 = self.RSS / (self.n - self.ENP)
        self.CCT = CCT * self.sigma2
        self.std_res = self.reside / (np.sqrt(self.sigma2 * (1.0 - self.diagonal)))
        self.bse = np.sqrt(self.CCT)
        self.cooksD = (self.std_res ** 2 * self.diagonal) / (self.ENP * (1.0 - self.diagonal))
        self.t_values = self.betas / self.bse
        self.df_model = self.n - self.tr_S
        self.df_reside = self.n - 2.0 * self.tr_S + self.tr_STS
        self.R2 = 1 - self.RSS / self.TSS
        self.adj_R2 = 1 - (1 - self.R2) * (self.n - 1) / (self.n - self.ENP - 1)
        self.llf = -np.log(self.RSS) * self.n / \
            2 - (1 + np.log(np.pi / self.n * 2)) * self.n / 2
        self.aic = -2.0 * self.llf + 2.0 * (self.tr_S + 1)
        self.aic_c = self.aic + 2.0 * self.tr_S * (self.tr_S + 1.0) / \
            (self.n - self.tr_S - 1.0)
        self.bic = -2.0 * self.llf + (self.k + 1) * np.log(self.n)

    def _build_wi(self, i, bw, tau):

        try:
            wi = Kernel(i, self.coords, self.t, bw, tau, fixed=self.fixed,
                        function=self.kernel).kernel
        except BaseException:
            raise  # TypeError('Unsupported kernel function  ', kernel)

        return wi

    def _local_fit(self, i, final=True, multi=False):
        wi = self._build_wi(i, self.bw, self.tau).reshape(-1, 1)
        betas, inv_xtx_xt = _compute_betas_gwr(self.y, self.X, wi)
        pre = np.dot(self.X[i], betas)[0]
        reside = self.y[i] - pre
        if multi:
            return betas.reshape(-1), reside
        influx = np.dot(self.X[i], inv_xtx_xt[:, i])
        if not final:
            return reside * reside, influx
        Si = np.dot(self.X[i], inv_xtx_xt).reshape(-1)
        CCT = np.diag(np.dot(inv_xtx_xt, inv_xtx_xt.T)).reshape(-1)
        Si2 = np.sum(Si ** 2)
        return influx, reside, pre, betas.reshape(-1), CCT, Si2

    def cal_aic(self):
        RSS = 0
        tr_S = 0
        aa = 0
        for i in range(self.n):
            err2, hat = self._local_fit(i, final=False)
            aa += err2 / ((1 - hat) ** 2)
            RSS += err2
            tr_S += hat
        llf = -np.log(RSS) * self.n / \
            2 - (1 + np.log(np.pi / self.n * 2)) * self.n / 2
        return CalAicObj(float(RSS), tr_S, float(llf), float(aa), self.n)

    def cal_multi(self):
        reside = np.empty((self.n, 1))
        betas = np.empty((self.n, self.k))
        for i in range(self.n):
            betas[i] = self._local_fit(i, multi=True)[0]
            reside[i] = self._local_fit(i, multi=True)[1]
        return CalMultiObj(betas, reside)


class MGTWR(GTWR):
    """
    Multiscale GTWR estimation and inference.

    Parameters
    ----------
    coords        : array-like
                    n*2, collection of n sets of (x,y) coordinates of
                    observatons

    t             : array
                    n*1, time location

    y             : array
                    n*1, dependent variable

    X             : array
                    n*k, independent variable, excluding the constant

    selector      : sel_bw object
                    valid sel_bw object that has successfully called
                    the "search" method. This parameter passes on
                    information from GAM model estimation including optimal
                    bandwidths.

    kernel        : string
                    type of kernel function used to weight observations;
                    available options:
                    'gaussian'
                    'bisquare'
                    'exponential'

    fixed         : boolean
                    True for distance based kernel function and  False for
                    adaptive (nearest neighbor) kernel function (default)

    constant      : boolean
                    True to include intercept (default) in model and False to exclude
                    intercept.
    after call fit() method
    Attributes
    ----------
    obj.betas           : array
                          n*k, estimated coefficients

    obj.pre             : array
                          n*1, predicted y values

    obj.CCT             : array
                          n*k, scaled variance-covariance matrix

    obj.df_model        : integer
                          model degrees of freedom

    obj.df_reside       : array
                          n*1, residuals of the response

    obj.RSS             : scalar
                          residual sum of squares

    obj.CCT             : array
                          n*k, scaled variance-covariance matrix

    obj.ENP             : scalar
                          effective number of parameters, which depends on
                          obj.sigma2

    obj.tr_S            : float
                          trace of S (hat) matrix

    obj.tr_STS          : float
                          trace of STS matrix

    obj.R2              : float
                          R-squared for the entire model (1- RSS/TSS)

    obj.adj_R2          : float
                          adjusted R-squared for the entire model

    obj.aic             : float
                          Akaike information criterion

    obj.aic_c           : float
                          corrected Akaike information criterion to account
                          to account for model complexity (smaller
                          bandwidths)

    obj.bic             : float
                          Bayesian information criterio

    obj.sigma2          : float
                          sigma squared (residual variance) that has been
                          corrected to account for the ENP

    obj.std_res         : array
                          n*1, standardised residuals

    obj.bse             : array
                          n*k, standard errors of parameters (betas)

    obj.t_values        : array
                          n*k, local t-statistics

    obj.llf             : scalar
                          log-likelihood of the full model; see
                          pysal.contrib.glm.family for damily-sepcific
                          log-likelihoods

    Examples
    --------
    import numpy as np
    from mgtwr.sel import SelBws
    from mgtwr.model import MGTWR
    np.random.seed(1)
    u = np.array([(i-1)%12 for i in range(1,1729)]).reshape(-1,1)
    v = np.array([((i-1)%144)//12 for i in range(1,1729)]).reshape(-1,1)
    t = np.array([(i-1)//144 for i in range(1,1729)]).reshape(-1,1)
    x1 = np.random.uniform(0,1,(1728,1))
    x2 = np.random.uniform(0,1,(1728,1))
    epsilon = np.random.randn(1728,1)
    beta0 = 5
    beta1 = 3 + (u + v + t)/6
    beta2 = 3+((36-(6-u)**2)*(36-(6-v)**2)*(36-(6-t)**2))/128
    y = beta0 + beta1 * x1 + beta2 * x2 + epsilon
    coords = np.hstack([u,v])
    X = np.hstack([x1,x2])
    sel_multi = SelBws(coords, t, X, y, kernel = 'gaussian', fixed = True, multi = True)
    sel_multi.search(multi_bw_min=[0.1],verbose=True,tol_multi=1.0e-4)
    mgtwr = MGTWR(coords, t, X, y, sel_multi, kernel = 'gaussian', fixed = True)
    mgtwr.fit()
    print(mgtwr.R2)
    0.9968863111153037
    """

    def __init__(
            self,
            coords: np.ndarray,
            t: np.ndarray,
            X: np.ndarray,
            y: np.ndarray,
            selector,
            kernel: str = 'bisquare',
            fixed: bool = False,
            constant: bool = True):
        self.selector = selector
        self.bws = self.selector.bws[0]  # final set of bandwidth
        self.taus = self.selector.bws[1]
        self.bw_ts = np.sqrt(self.bws ** 2 / self.taus)
        self.bws_history = selector.bws[2]  # bws history in back_fitting
        self.taus_history = selector.bws[3]
        self.betas = selector.bws[5]
        bw_init = self.selector.bws[7]  # initialization bandwidth
        tau_init = self.selector.bws[8]
        super().__init__(coords, t, X, y, bw_init, tau_init,
                         kernel=kernel, fixed=fixed, constant=constant)
        self.n_chunks = None
        self.ENP_j = None

    def _chunk_compute(self, chunk_id=0):

        n = self.n
        k = self.k
        n_chunks = self.n_chunks
        chunk_size = int(np.ceil(float(n / n_chunks)))
        ENP_j = np.zeros(self.k)
        CCT = np.zeros((self.n, self.k))

        chunk_index = np.arange(n)[chunk_id * chunk_size:(chunk_id + 1) * chunk_size]
        init_pR = np.zeros((n, len(chunk_index)))
        init_pR[chunk_index, :] = np.eye(len(chunk_index))
        pR = np.zeros((n, len(chunk_index),
                       k))  # partial R: n by chunk_size by k

        for i in range(n):
            wi = self._build_wi(i, self.bw, self.tau).reshape(-1, 1)
            xT = (self.X * wi).T
            P = np.linalg.solve(xT.dot(self.X), xT).dot(init_pR).T
            pR[i, :, :] = P * self.X[i]

        err = init_pR - np.sum(pR, axis=2)  # n by chunk_size

        for iter_i in range(self.bws_history.shape[0]):
            for j in range(k):
                pRj_old = pR[:, :, j] + err
                Xj = self.X[:, j]
                n_chunks_Aj = n_chunks
                chunk_size_Aj = int(np.ceil(float(n / n_chunks_Aj)))
                for chunk_Aj in range(n_chunks_Aj):
                    chunk_index_Aj = np.arange(n)[chunk_Aj * chunk_size_Aj:(
                                                                                   chunk_Aj + 1) * chunk_size_Aj]
                    pAj = np.empty((len(chunk_index_Aj), n))
                    for i in range(len(chunk_index_Aj)):
                        index = chunk_index_Aj[i]
                        wi = self._build_wi(index, self.bws_history[iter_i, j],
                                            self.taus_history[iter_i, j])
                        xw = Xj * wi
                        pAj[i, :] = Xj[index] / np.sum(xw * Xj) * xw
                    pR[chunk_index_Aj, :, j] = pAj.dot(pRj_old)
                err = pRj_old - pR[:, :, j]

        for j in range(k):
            CCT[:, j] += ((pR[:, :, j] / self.X[:, j].reshape(-1, 1)) ** 2).sum(
                axis=1)
        for i in range(len(chunk_index)):
            ENP_j += pR[chunk_index[i], i, :]

        return ENP_j, CCT,

    def fit(self, n_chunks=1):
        """
        Compute MGTWR inference by chunk to reduce memory footprint.
        """
        self.n_chunks = n_chunks
        betas = self.selector.bws[5]
        pre = np.sum(self.X * betas, axis=1).reshape(-1, 1)
        result = map(self._chunk_compute, (range(n_chunks)))
        result_list = list(zip(*result))
        ENP_j = np.sum(np.array(result_list[0]), axis=0)
        CCT = np.sum(np.array(result_list[1]), axis=0)
        self.pre = pre
        self.betas = betas
        self.ENP_j = ENP_j
        self.reside = self.y - self.pre
        self.tr_S = np.sum(self.ENP_j)
        self.ENP = self.tr_S
        self.TSS = np.sum((self.y - np.mean(self.y)) ** 2)
        self.RSS = np.sum(self.reside ** 2)
        self.sigma2 = self.RSS / (self.n - self.tr_S)
        self.CCT = CCT * self.sigma2
        self.bse = np.sqrt(self.CCT)
        self.t_values = self.betas / self.bse
        self.df_model = self.n - self.tr_S
        self.R2 = 1 - self.RSS / self.TSS
        self.adj_R2 = 1 - (1 - self.R2) * (self.n - 1) / (self.n - self.ENP - 1)
        self.llf = -np.log(self.RSS) * self.n / \
            2 - (1 + np.log(np.pi / self.n * 2)) * self.n / 2
        self.aic = -2.0 * self.llf + 2.0 * (self.tr_S + 1)
        self.aic_c = self.aic + 2.0 * self.tr_S * (self.tr_S + 1.0) / \
            (self.n - self.tr_S - 1.0)
        self.bic = -2.0 * self.llf + (self.k + 1) * np.log(self.n)



import numpy as np


def get_AICc(gtwr):
    """
    Get AICc value

    Gaussian: p61, (2.33), Fotheringham, Brunsdon and Charlton (2002)

    GWGLM: $AICc=AIC+2k(k+1)/(n-k-1)$, Nakaya et al. (2005): p2704, (36)

    """
    n = gtwr.n
    k = gtwr.tr_S

    aic_c = get_AIC(gtwr) + 2.0 * k * (k + 1.0) / (n - k - 1.0)
    return aic_c


def get_AIC(gtwr):
    """
    Get AIC calue

    Gaussian: p96, (4.22), Fotheringham, Brunsdon and Charlton (2002)

    GWGLM:  AIC(G)=D(G) + 2K(G), where D and K denote the deviance and the effective
    number of parameters in the model with bandwidth G, respectively.

    """

    k = gtwr.tr_S

    aic = -2.0 * gtwr.llf + 2.0 * (k + 1)

    return aic


def get_BIC(gtwr):
    """
    Get BIC value

    Gaussian: p61 (2.34), Fotheringham, Brunsdon and Charlton (2002)
    $BIC = -2log(L)+klog(n)$

    GWGLM: $BIC = dev + tr_S * log(n)$

    """
    n = gtwr.n  # (scalar) number of observations
    k = gtwr.tr_S

    bic = -2.0 * gtwr.llf + (k + 1) * np.log(n)
    return bic


def get_CV(gtwr):
    """
    Get CV value

    Gaussian only

    Methods: p60, (2.31) or p212 (9.4)
    Fotheringham, A. S., Brunsdon, C., & Charlton, M. (2002).
    Geographically weighted regression: the analysis of spatially varying relationships.
    Modification: sum of residual squared is divided by n according to GWR4 results

    """

    cv = gtwr.aa / gtwr.n
    return cv


def corr(cov):
    inv_sd = np.diag(1 / np.sqrt(np.diag(cov)))
    cor_s = np.dot(np.dot(inv_sd, cov), inv_sd)
    return cor_s



import numpy as np
from .diagnosis import get_AICc, get_AIC, get_BIC, get_CV
# from scipy.spatial.distance import pdist
from .model import GTWR
from .search import two_d_golden_section, multi_bws

get_diagnosis = {'AICc': get_AICc, 'AIC': get_AIC, 'BIC': get_BIC, 'CV': get_CV}


class SelBws:

    def __init__(
            self,
            coords: np.ndarray,
            t: np.ndarray,
            X: np.ndarray,
            y: np.ndarray,
            kernel: str = 'bisquare',
            fixed: bool = True,
            multi: bool = False,
            constant: bool = True
    ):
        """
        Select bandwidth for kernel

        Parameters
        ----------
        coords        : array-like
                        n*2, collection of n sets of (x,y) coordinates of
                        observatons

        t             : array
                        n*1, time location

        y             : array
                        n*1, dependent variable

        X             : array
                        n*k, independent variable, excluding the constant

        kernel        : string
                        type of kernel function used to weight observations;
                        available options:
                        'gaussian'
                        'bisquare'
                        'exponential'

        fixed         : boolean
                        True for distance based kernel function and  False for
                        adaptive (nearest neighbor) kernel function (default)

        constant      : boolean
                        True to include intercept (default) in model and False to exclude
                        intercept.

        multi          : True for multiple (covaraite-specific) bandwidths
                         False for a traditional (same for  all covariates)
                         bandwidth; default is False.

        Examples
        --------
        import numpy as np
        from mgtwr.sel import SelBws
        from mgtwr.model import GTWR, MGTWR
        np.random.seed(1)
        u = np.array([(i-1)%12 for i in range(1,1729)]).reshape(-1,1)
        v = np.array([((i-1)%144)//12 for i in range(1,1729)]).reshape(-1,1)
        t = np.array([(i-1)//144 for i in range(1,1729)]).reshape(-1,1)
        x1 = np.random.uniform(0,1,(1728,1))
        x2 = np.random.uniform(0,1,(1728,1))
        epsilon = np.random.randn(1728,1)
        beta0 = 5
        beta1 = 3 + (u + v + t)/6
        beta2 = 3+((36-(6-u)**2)*(36-(6-v)**2)*(36-(6-t)**2))/128
        y = beta0 + beta1 * x1 + beta2 * x2 + epsilon
        coords = np.hstack([u,v])
        X = np.hstack([x1,x2])
        sel = SelBws(coords, t, X, y, kernel = 'gaussian', fixed = True)
        bw, tau = sel.search(bw_max = 40, tau_max = 5, verbose = True)
        0.8, 1.1
        """
        self.coords = coords
        self.t = t
        self.n = X.shape[0]
        if constant:
            self.X = np.hstack([np.ones((self.n, 1)), X])
        else:
            self.X = X
        self.k = self.X.shape[1]
        self.y = y
        self.constant = constant
        self.kernel = kernel
        self.fixed = fixed
        self.multi = multi
        self.int_score = not self.fixed
        self.criterion = None
        self.bws = None
        self.tol = None
        self.bw_decimal = None
        self.tau_decimal = None

    def search(
            self,
            criterion: str = 'AICc',
            bw_min: float = None,
            bw_max: float = None,
            tau_min: float = None,
            tau_max: float = None,
            tol: float = 1.0e-6,
            bw_decimal: int = 1,
            tau_decimal: int = 1,
            init_bw: float = None,
            init_tau: float = None,
            multi_bw_min: list = None,
            multi_bw_max: list = None,
            multi_tau_min: list = None,
            multi_tau_max: list = None,
            tol_multi: float = 1.0e-5,
            verbose: bool = False,
            rss_score: bool = False
    ):
        """
        Method to select one unique bandwidth and Spatio-temporal scale for a gtwr model or a
        bandwidth vector and Spatio-temporal scale vector for a mtgwr model.

        Parameters
        ----------
        criterion      : string
                         bw selection criterion: 'AICc', 'AIC', 'BIC', 'CV'
        bw_min         : float
                         min value used in bandwidth search
        bw_max         : float
                         max value used in bandwidth search
        tau_min        : float
                         min value used in spatio-temporal scale search
        tau_max        : float
                         max value used in spatio-temporal scale search
        multi_bw_min   : list
                         min values used for each covariate in mgwr bandwidth search.
                         Must be either a single value or have one value for
                         each covariate including the intercept
        multi_bw_max   : list
                         max values used for each covariate in mgwr bandwidth
                         search. Must be either a single value or have one value
                         for each covariate including the intercept
        multi_tau_min  : list
                         min values used for each covariate in mgtwr spatio-temporal scale
                         search. Must be either a single value or have one value
                         for each covariate including the intercept
        multi_tau_max  : max values used for each covariate in mgtwr spatio-temporal scale
                         search. Must be either a single value or have one value
                         for each covariate including the intercept
        tol            : float
                         tolerance used to determine convergence
        bw_decimal     : int
                        The number of bw decimal places reserved
        tau_decimal    : int
                        The number of tau decimal places reserved
        init_bw        : float
                         None (default) to initialize MGTWR with a bandwidth
                         derived from GTWR. Otherwise this option will choose the
                         bandwidth to initialize MGWR with.
        init_tau       : float
                         None (default) to initialize MGTWR with a spatio-temporal scale
                         derived from GTWR. Otherwise this option will choose the
                         spatio-temporal scale to initialize MGWR with.
        tol_multi      : convergence tolerance for the multiple bandwidth
                         back fitting algorithm; a larger tolerance may stop the
                         algorithm faster though it may result in a less optimal
                         model
        rss_score      : True to use the residual sum of squares to evaluate
                         each iteration of the multiple bandwidth back fitting
                         routine and False to use a smooth function; default is
                         False
        verbose        : Boolean
                         If true, bandwidth searching history is printed out; default is False.
        """
        self.criterion = criterion
        self.tol = tol
        self.bw_decimal = bw_decimal
        self.tau_decimal = tau_decimal
        if self.multi:
            if multi_bw_min is not None:
                if len(multi_bw_min) == self.k:
                    multi_bw_min = multi_bw_min
                elif len(multi_bw_min) == 1:
                    multi_bw_min = multi_bw_min * self.k
                else:
                    raise AttributeError(
                        "multi_bw_min must be either a list containing"
                        " a single entry or a list containing an entry for each of k"
                        " covariates including the intercept")
            else:
                a = self._init_section(bw_min, bw_max, tau_min, tau_max)[0]
                multi_bw_min = [a] * self.k

            if multi_bw_max is not None:
                if len(multi_bw_max) == self.k:
                    multi_bw_max = multi_bw_max
                elif len(multi_bw_max) == 1:
                    multi_bw_max = multi_bw_max * self.k
                else:
                    raise AttributeError(
                        "multi_bw_max must be either a list containing"
                        " a single entry or a list containing an entry for each of k"
                        " covariates including the intercept")
            else:
                c = self._init_section(bw_min, bw_max, tau_min, tau_max)[1]
                multi_bw_max = [c] * self.k

            if multi_tau_min is not None:
                if len(multi_tau_min) == self.k:
                    multi_tau_min = multi_tau_min
                elif len(multi_tau_min) == 1:
                    multi_tau_min = multi_tau_min * self.k
                else:
                    raise AttributeError(
                        "multi_tau_min must be either a list containing"
                        " a single entry or a list containing an entry for each of k"
                        " variates including the intercept")
            else:
                A = self._init_section(bw_min, bw_max, tau_min, tau_max)[2]
                multi_tau_min = [A] * self.k

            if multi_tau_max is not None:
                if len(multi_tau_max) == self.k:
                    multi_tau_max = multi_tau_max
                elif len(multi_tau_max) == 1:
                    multi_tau_max = multi_tau_max * self.k
                else:
                    raise AttributeError(
                        "multi_tau_max must be either a list containing"
                        " a single entry or a list containing an entry for each of k"
                        " variates including the intercept")
            else:
                C = self._init_section(bw_min, bw_max, tau_min, tau_max)[3]
                multi_tau_max = [C] * self.k

        if self.multi:
            self.bws = multi_bws(init_bw, init_tau, self.X, self.y, self.n, self.k, tol_multi,
                                 rss_score, self.gtwr_func, self.bw_func, self.sel_func, multi_bw_min, multi_bw_max,
                                 multi_tau_min, multi_tau_max, verbose=verbose)
            bw = self.bws[0]
            tau = self.bws[1]
            return bw, tau
        else:
            bw, tau = self._bw(bw_min, bw_max, tau_min, tau_max, tol, bw_decimal, tau_decimal, verbose)
            return bw, tau

    def _bw(self, bw_min, bw_max, tau_min, tau_max, tol, bw_decimal, tau_decimal, verbose):
        bw_min, bw_max, tau_min, tau_max = self._init_section(bw_min, bw_max, tau_min, tau_max)
        delta = 0.38197  # 1 - (np.sqrt(5.0)-1.0)/2.0
        bw, tau = two_d_golden_section(bw_min, bw_max, tau_min, tau_max, delta, self.gwr_func, tol,
                                       bw_decimal, tau_decimal, verbose)
        return bw, tau

    def gwr_func(self, bw, tau):
        return get_diagnosis[self.criterion](GTWR(
            self.coords, self.t, self.X, self.y, bw, tau, kernel=self.kernel,
            fixed=self.fixed, constant=False).cal_aic())

    def gtwr_func(self, X, y, bw, tau):
        return GTWR(self.coords, self.t, X, y, bw, tau, kernel=self.kernel,
                    fixed=self.fixed, constant=False).cal_multi()

    def bw_func(self, X, y):
        selector = SelBws(self.coords, self.t, X, y, kernel=self.kernel, fixed=self.fixed,
                          constant=False)
        return selector

    def sel_func(self, bw_func, bw_min=None, bw_max=None, tau_min=None, tau_max=None):
        return bw_func.search(criterion=self.criterion, bw_min=bw_min, bw_max=bw_max, tau_min=tau_min, tau_max=tau_max,
                              tol=self.tol, bw_decimal=self.bw_decimal, tau_decimal=self.tau_decimal, verbose=False)

    def _init_section(self, bw_min, bw_max, tau_min, tau_max):
        # if len(X) > 0:
        #     n_glob = X.shape[1]
        # else:
        #     n_glob = 0
        # if constant:
        #     n_vars = n_glob + 1
        # else:
        #     n_vars = n_glob
        # n = np.array(coords).shape[0]

        # if self.int_score:
        #     a = 40 + 2 * n_vars
        #     c = n
        # else:
        #     sq_dists = pdist(coords)
        #     a = np.min(sq_dists) / 2.0
        #     c = np.max(sq_dists) * 2.0

        # if self.bw_min is not None:
        #     a = self.bw_min
        # if self.bw_max is not None:
        #     c = self.bw_max

        # if self.tau_min is not None:
        #     A = self.tau_min
        # else:
        #     A = 0
        # if self.tau_max is not None:
        #     C = self.tau_max
        # else:
        #     C = 2
        a = bw_min if bw_min is not None else 0
        if bw_max is not None:
            c = bw_max
        else:
            c = max(np.max(self.coords[:, 0]) - np.min(self.coords[:, 0]),
                    np.max(self.coords[:, 1]) - np.min(self.coords[:, 1]))

        A = tau_min if tau_min is not None else 0
        C = tau_max if tau_max is not None else 4

        return a, c, A, C



import numpy as np
from copy import deepcopy


def golden_section(A, C, x, delta, tau_decimal, function, tol):
    iter_num = 0
    score_dict = {}
    diff = 1e9
    opt_score = None
    opt_tau = None
    B = A + delta * np.abs(C - A)
    D = C - delta * np.abs(C - A)
    while np.abs(diff) > tol and iter_num < 200:
        iter_num += 1
        B = np.round(B, tau_decimal)
        D = np.round(D, tau_decimal)
        if B in score_dict:
            score_B = score_dict[B]
        else:
            score_B = function(x, B)
            score_dict[B] = score_B

        if D in score_dict:
            score_D = score_dict[D]
        else:
            score_D = function(x, D)
            score_dict[D] = score_D
        if score_B <= score_D:
            opt_score = score_B
            opt_tau = B
            C = D
            D = B
            B = A + delta * np.abs(C - A)
        else:
            opt_score = score_D
            opt_tau = D
            A = B
            B = D
            D = C - delta * np.abs(C - A)
        diff = score_B - score_D
    return opt_tau, opt_score


def two_d_golden_section(a, c, A, C, delta, function,
                         tol, bw_decimal, tau_decimal, verbose=False):
    b = a + delta * np.abs(c - a)
    d = c - delta * np.abs(c - a)
    opt_bw = None
    opt_tau = None
    diff = 1e9
    score_dict = {}
    iter_num = 0
    while np.abs(diff) > tol and iter_num < 200:
        iter_num += 1
        b = np.round(b, bw_decimal)
        d = np.round(d, bw_decimal)
        if b in score_dict:
            tau_b, score_b = score_dict[b]
        else:
            tau_b, score_b = golden_section(A, C, b, delta, tau_decimal, function, tol)
            score_dict[b] = [tau_b, score_b]
        if d in score_dict:
            tau_d, score_d = score_dict[d]
        else:
            tau_d, score_d = golden_section(A, C, d, delta, tau_decimal, function, tol)
            score_dict[d] = [tau_d, score_d]
        if score_b <= score_d:
            opt_score = score_b
            opt_bw = b
            opt_tau = tau_b
            c = d
            d = b
            b = a + delta * np.abs(c - a)
        else:
            opt_score = score_d
            opt_bw = d
            opt_tau = tau_d
            a = b
            b = d
            d = c - delta * np.abs(c - a)
        diff = score_b - score_d
        if verbose:
            print('----------\n bw:%s, tau:%s, score:%s' % (opt_bw, opt_tau, opt_score))
    return opt_bw, opt_tau


def multi_bws(init_bw, init_tau, X, y, n, k, tol, rss_score,
              gtwr_func, bw_func, sel_func, multi_bw_min, multi_bw_max,
              multi_tau_min, multi_tau_max, verbose=False):
    """
    Multiscale GTWR bandwidth search procedure using iterative GAM back fitting
    """
    if init_bw or init_tau is None:
        bw, tau = sel_func(bw_func(X, y))
    else:
        bw, tau = init_bw, init_tau
    opt_model = gtwr_func(X, y, bw, tau)
    bw_gtwr = bw
    tau_gtwr = tau
    err = opt_model.reside
    betas = opt_model.betas

    XB = np.multiply(betas, X)
    rss = np.sum(err ** 2) if rss_score else None
    scores = []
    bws = np.empty(k)
    taus = np.empty(k)
    BWs = []
    Taus = []
    Betas = None

    for iter_num in range(1, 201):
        new_XB = np.zeros_like(X)
        Betas = np.zeros_like(X)

        for j in range(k):
            temp_y = XB[:, j].reshape((-1, 1))
            temp_y = temp_y + err
            temp_X = X[:, j].reshape((-1, 1))
            bw_class = bw_func(temp_X, temp_y)

            bw, tau = sel_func(bw_class, multi_bw_min[j], multi_bw_max[j],
                               multi_tau_min[j], multi_tau_max[j])

            opt_model = gtwr_func(temp_X, temp_y, bw, tau)
            err = opt_model.reside
            betas = opt_model.betas
            new_XB[:, j] = (betas * temp_X).reshape(-1)
            Betas[:, j] = betas.reshape(-1)
            bws[j] = bw
            taus[j] = tau

        num = np.sum((new_XB - XB) ** 2) / n
        den = np.sum(np.sum(new_XB, axis=1) ** 2)
        score = (num / den) ** 0.5
        XB = new_XB

        if rss_score:
            predy = np.sum(np.multiply(betas, X), axis=1).reshape((-1, 1))
            new_rss = np.sum((y - predy) ** 2)
            score = np.abs((new_rss - rss) / new_rss)
            rss = new_rss
        scores.append(deepcopy(score))
        delta = score
        BWs.append(deepcopy(bws))
        Taus.append(deepcopy(taus))

        if verbose:
            print("Current iteration:", iter_num, ",SOC:", np.round(score, 7))
            print("Bandwidths:", ', '.join([str(bw) for bw in bws]))
            print("taus:", ','.join([str(tau) for tau in taus]))

        if delta < tol:
            break
    opt_bws = BWs[-1]
    opt_tau = Taus[-1]
    return (opt_bws, opt_tau, np.array(BWs), np.array(Taus), np.array(scores),
            Betas, err, bw_gtwr, tau_gtwr)




class CalAicObj:

    def __init__(self, RSS, tr_S, llf, aa, n):
        self.RSS = RSS
        self.tr_S = tr_S
        self.llf = llf
        self.aa = aa
        self.n = n


class CalMultiObj:

    def __init__(self, betas, reside):
        self.betas = betas
        self.reside = reside
